import datetime as dt

__title__ = "gdxpds"
__description__ = "gdx-pandas is a python package to translate between gdx (GAMS data) and pandas"
__url__ = "https://github.com/NREL/gdx-pandas"
__version__ = "1.4.0"
__author__ = "Elaine T. Hale"
__author_email__ = "elaine.hale@nrel.gov"
__license__ = "BSD-3"
__copyright__ = "Copyright {}, The Alliance for Sustainable Energy, LLC".format(
    dt.date.today().year
)
